! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      use crlibm_lib
      
      implicit none
      
      ! Starting specifics for loading in a radiative envelope mixing profile.
      ! Remember to set the path to the loaded mixing profile. First column has to be the q values (m/Mstar),
      ! and second column must be the Dmix values to be used instead.
      ! Attached is an example profile which can be loaded in. It is very similar to the one of IGWs used
      ! in Pedersen et al. (2018), and is basically a scaled rho^-1/2.
      character (len=*), PARAMETER :: FileName = 'PathToRadiativeEnvelopeMixingProfile/InverseSqrtRho.dat'
      double precision, allocatable :: qIN(:)                      
      double precision, allocatable :: DmixIN(:)

      contains
      

      
      subroutine extras_controls(id, ierr)
        integer, intent(in) :: id
        integer, intent(out) :: ierr
        type (star_info), pointer :: s
        integer::narg,cptArg
        ierr = 0
        call star_ptr(id, s, ierr)
        if (ierr /= 0) return
        
        write(*,*)id
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
         
         ! Uncomment these lines if you wish to use the functions in this file,
         ! otherwise we use a null_ version which does nothing.
         s% extras_startup => extras_startup
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns 
        
         ! Point to the new D_mix profile to be computed
         s% other_D_mix => my_other_D_mix

         ! Once you have set the function pointers you want,
         ! then uncomment this (or set it in your star_job inlist)
         ! to disable the printed warning message,
          s% job% warn_run_star_extras =.false.     
            
      end subroutine extras_controls
      
      ! None of the following functions are called unless you set their
      ! function point in extras_control.
      
      
      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: i, nlines
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_startup = 0
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
         
         ! Count number of rows in the datafile including the shape of the
         ! radiative envelope mixing
         nlines = 0 
         OPEN (1, file = FileName) 
         DO 
            READ (1,*, END=10) 
            nlines = nlines + 1 
         END DO 
         10 CLOSE (1) 

        print*, nlines
        
        ! Allocate space for the two parameters in the datafile,
        ! i.e. q and Dmix
        allocate(qIN(nlines))
        allocate(DmixIN(nlines))
         
         ! Read in the input Dmix profile
            open(unit=20,file=FileName,status='old')
            do i=1,size(qIN) ! Read in values until end of the length of the array
                read(20,*) qIN(i), DmixIN(i)
            end do
            close(20)

            
      end function extras_startup
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going   
         
         ! The following shouldn't really be necessary and can be left out


         ! if you want to check multiple conditions, it can be useful
         ! to set a different termination code depending on which
         ! condition was triggered.  MESA provides 9 customizeable
         ! termination codes, named t_xtra1 .. t_xtra9.  You can
         ! customize the messages that will be printed upon exit by
         ! setting the corresponding termination_code_str value.
         ! termination_code_str(t_xtra1) = 'my termination condition'

         ! by default, indicate where (in the code) MESA terminated
         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
      end function extras_check_model


      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s         
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         !note: do NOT add the extras names to history_columns.list
         ! the history_columns.list is only for the built-in log column options.
         ! it must not include the new column names you are adding here.       

      end subroutine data_for_extra_history_columns

      subroutine my_other_D_mix(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         real(dp) :: OldYMax, OldYMin, OldYRange, NewYMax, NewYMin, NewYRange
         real(dp) :: NewDmixIN(size(qIN))
         real(dp) :: OldXMax, OldXMin, OldXRange, NewXMax, NewXMin, NewXRange
         real(dp) :: NewqIN(size(qIN))
         real(dp) :: ydiff, xdiff, division
         integer :: k, ii, idx
         integer :: q_DmixMin_New, D_switch   
         
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return        
                 
         
        ! If none of these logical statements are true, MESA will use the implemented minimum mixing
        ! descriptions instead of the loaded profile.            
         if ((.not. s% doing_first_model_of_run) .and. (.not. s% doing_relax )) then   ! takes care of missing startup values         
         
         
         !--------------------------------------------------------------------------------------------
         !---------------- Determining position of switch to loaded profile --------------------------
         !--------------------------------------------------------------------------------------------
         ! The switch to the loaded mixing profile is set to occur when the Dmix value becomes
         ! less or equal to the min_D_mix parameter when going from the center of the star towards
         ! the surface.
         D_switch = s% min_D_mix 
         do k = 1, s% nz
            idx = s% nz - k 
            if (s% D_mix(idx) <= D_switch) then
                q_DmixMin_New = idx 
                exit
            end if
         end do   

         
         !--------------------------------------------------------------------------------------------
         !------------------------------- Rescaling Y-axis -------------------------------------------
         !--------------------------------------------------------------------------------------------
         ! Rescale the y-axis of the loaded Dmix profile according to the min_Dmix.
         ! The value for the s% min_Dmix is forced on the innermost point of the loaded Dmix
         ! profile, i.e. the point closest to the center
         do ii = 1, size(qIN)
            NewDmixIN(ii) = DmixIN(ii)/DmixIN(size(qIN)) *  s% min_D_mix
         end do

         if (q_DmixMin_New > 0) then
         !--------------------------------------------------------------------------------------------
         !------------------------------- Rescaling X-axis -------------------------------------------
         !--------------------------------------------------------------------------------------------
         ! I.e. rescale qIN for the loaded Dmix profile
         OldXMax = 1
         OldXMin = minval(qIN)
         NewXMax = 1
         NewXMin = s% q(q_DmixMin_New)
         OldXRange = OldXMax - OldXMin
         NewXRange = NewXMax - NewXMin
         do ii = 1, size(qIN)
            NewqIN(ii) = (((qIN(ii) - OldXMin) * NewXRange) / OldXRange) + NewXMin
         end do         

         
         !--------------------------------------------------------------------------------------------
         !------------------------------- Do a linear interpolation ----------------------------------
         !--------------------------------------------------------------------------------------------
         do ii = 2, size(qIN)
            do k = 1, q_DmixMin_New
                s% mixing_type(k) = 8               
                if ( (s%q(k) <= NewqIN(ii-1)) .and. (s%q(k) >= NewqIN(ii)) ) then
                    ydiff = NewDmixIN(ii-1) - NewDmixIN(ii)
                    xdiff = s%q(k) - NewqIN(ii)
                    division = NewqIN(ii-1) - NewqIN(ii)
                    s% D_mix(k) = NewDmixIN(ii) + ydiff*xdiff/division
                end if
            end do
         end do
         
         
         !--------------------------------------------------------------------------------------------
         !------------------------------- Finished rescaling -----------------------------------------
         !--------------------------------------------------------------------------------------------          
         ! Loaded Dmix profile will not be implemented if MESA is doing the startup of a run  
         ! or relaxing the starting model
         end if
         else 
           write(*,*) 'doing_first_model_of_run or relaxing starting model ----> not computing envelope D_mix'
         end if
         
         

      end subroutine my_other_D_mix
      
      integer function how_many_extra_profile_columns(id, id_extra)
         use star_def, only: star_info
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         !note: do NOT add the extra names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.
         
         
         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'
 
         
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      
      integer function extras_finish_step(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr, ii, slen
         !real(16) :: dxc_tolerance, delta_xc, delta_age, xc_last=-99.0d0, age_last=-99.d0
         double precision :: save_points(63)
         character(len=3) :: hval(63)
         character(len=300) :: profname
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going
         call store_extra_info(s)
         

         ! see extras_check_model for information about custom termination codes
         ! by default, indicate where (in the code) MESA terminated
         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step
      

      
      subroutine extras_after_evolve(id, id_extra, ierr)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine extras_after_evolve
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      end module run_star_extras
      
